/*
    Kiwi
    Bitboard management

    Copyright (c) 1999-2004,2005 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef BITBOARD_H_
#define BITBOARD_H_

// Note: define USE_ASM and a target platform (WIN_I386 or LINUX_I386)
// to enable inline assembly. This will make move generation considerably faster,
// though in practice it doesn't influence playing strength too much.

#include <stdio.h>
#include "platform.h"

const int BitCount[256] = {
    0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
};

const int FirstBit[256] = {
   -1, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    7, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
};

/**
    Returns the index of the first bit set in data, or -1 if no bit is set.
*/
inline int bitScanForward( Uint64 data )
{
#if defined(USE_ASM)
#if defined(LINUX_I386) || defined(WIN_I386)
#if defined(_MSC_VER)
#pragma warning(disable: 4035)
    // Microsoft Visual C++ inline assembly for I386
    __asm {
        mov     edx, DWORD PTR [data + 0]
        test    edx, edx
         jz     High_
        bsf     eax, edx
         jmp    Exit_
    High_:
        mov     edx, DWORD PTR [data + 4]
        mov     eax, -1
        test    edx, edx
         jz     Exit_
        bsf     eax, edx
        add     eax, 32
    Exit_:
    }
#elif defined(__GNUC__)
    // GCC inline assembly for I386
    int bit;

    asm(
        "   testl   %%eax, %%eax    \n\t"
        "    jz     1f              \n\t"
        "   bsf     %%eax, %0       \n\t"
        "    jmp    2f              \n\t"
        "1:                         \n\t"
        "   movl    $-1, %0         \n\t"
        "   testl   %%edx, %%edx    \n\t"
        "    jz     2f              \n\t"
        "   bsf     %%edx, %0       \n\t"
        "   addl    $32, %0         \n\t"
        "2:                         \n\t"
        : "=&r" (bit)
        : "A" (data)
    );

    return bit;
#else
#error Unknown compiler for 32-bit I386 platform
#endif // Compiler selection
#endif // I386 platform
#else // Do not use assembly
    int bit;

    if( data != 0 ) {
        Uint32 x;

        if( data & 0xFFFFFFFF ) {
            x = (Uint32) (data & 0xFFFFFFFF);
            bit = 0;
        }
        else {
            x = (Uint32) (data >> 32);
            bit = 32;
        }

        if( x & 0xFF ) {
            bit += FirstBit[ x & 0xFF ];
        }
        else if( x & 0xFF00 ) {
            bit += 8 + FirstBit[ (x >> 8) & 0xFF ];
        }
        else if( x & 0xFF0000 ) {
            bit += 16 + FirstBit[ (x >> 16) & 0xFF ];
        }
        else {
            bit += 24 + FirstBit[ (x >> 24) & 0xFF ];
        }
    }
    else {
        bit = -1;
    }

    return bit;
#endif
}

/**
    Returns the number of bit set in data.
*/
inline int bitCount( Uint64 data )
{
#if defined(USE_ASM)
#if defined(LINUX_I386) || defined(WIN_I386)
#if defined(_MSC_VER)
#pragma warning(disable: 4035)
    // Microsoft Visual C++ inline assembly for I386
    __asm {
            mov     edx, DWORD PTR [data + 0]
            xor     eax, eax
            test    edx, edx
             jz     DoneLo
        TestLo:
            inc     eax
            lea     ecx, [edx-1]
            and     edx, ecx
             jnz    TestLo
        DoneLo:
            mov     edx, DWORD PTR [data + 4]
            test    edx, edx
             jz     DoneHi
        TestHi:
            inc     eax
            lea     ecx, [edx-1]
            and     edx, ecx
             jnz    TestHi
        DoneHi:
    }
#elif defined(__GNUC__)
    // GCC inline assembly for I386
    int count;

    // Note: we must preserve EAX and EDX here
    asm(
        "   xor     %0, %0          \n\t"
        "   testl   %%eax, %%eax    \n\t"
        "    jz     1f              \n\t"
        "   push    %%eax           \n\t"
        "0:                         \n\t"
        "   inc     %0              \n\t"
        "   lea     -1(%%eax),%%ecx \n\t"
        "   and     %%ecx, %%eax    \n\t"
        "    jnz    0b              \n\t"
        "   pop     %%eax           \n\t"
        "1:                         \n\t"
        "   testl   %%edx, %%edx    \n\t"
        "    jz     3f              \n\t"
        "   push    %%edx           \n\t"
        "2:                         \n\t"
        "   inc     %0              \n\t"
        "   lea     -1(%%edx),%%ecx \n\t"
        "   and     %%ecx, %%edx    \n\t"
        "    jnz    2b              \n\t"
        "   pop     %%edx           \n\t"
        "3:                         \n\t"
        : "=&r" (count)
        : "A" (data)
        : "%ecx"
    );

    return count;
#else
#error Unknown compiler for 32-bit I386 platform
#endif // Compiler selection
#endif // I386 platform
#else // Do not use assembly
    int result;

    result =
        BitCount[ ((data >> 56) & 0xFF) ] +
        BitCount[ ((data >> 48) & 0xFF) ] +
        BitCount[ ((data >> 40) & 0xFF) ] +
        BitCount[ ((data >> 32) & 0xFF) ] +
        BitCount[ ((data >> 24) & 0xFF) ] +
        BitCount[ ((data >> 16) & 0xFF) ] +
        BitCount[ ((data >>  8) & 0xFF) ] +
        BitCount[ ((data      ) & 0xFF) ];

    return result;
#endif
}

/**
    Bitboard, i.e. a 64-bit unsigned integer with some extra chess-related functions.
*/
class BitBoard
{
public:
    static const BitBoard   Clr[64];
    static const BitBoard   Set[64];
    static const BitBoard   Zero;

    //
    BitBoard() {
    }

    BitBoard( Uint32 lo, Uint32 hi ) {
        data = ((Uint64)hi << 32) | ((Uint64)lo);
    }

    BitBoard( Uint64 n ) {
        data = n;
    }

    BitBoard( const BitBoard & bb ) {
        data = bb.data;
    }

    //
    ~BitBoard() {
    }

    //
    void operator = ( const BitBoard & bb ) {
        data = bb.data;
    }

    void operator = ( Uint64 bb ) {
        data = bb;
    }

    //
    void operator &= ( const BitBoard & bb ) {
        data &= bb.data;
    }

    void operator |= ( const BitBoard & bb ) {
        data |= bb.data;
    }

    void operator ^= ( const BitBoard & bb ) {
        data ^= bb.data;
    }

    //
    void operator &= ( Uint64 bb ) {
        data &= bb;
    }

    void operator |= ( Uint64 bb ) {
        data |= bb;
    }

    void operator ^= ( Uint64 bb ) {
        data ^= bb;
    }

    //
    void operator <<= ( int len ) {
        data <<= len;
    }

    void operator >>= ( int len ) {
        data >>= len;
    }

    //
    BitBoard operator << ( int len ) const {
        return BitBoard( data << len );
    }

    BitBoard operator >> ( int len ) const {
        return BitBoard( data >> len );
    }

    //
    BitBoard operator ~ () const {
        return BitBoard( ~data );
    }

    BitBoard operator & ( const BitBoard & bb ) const {
        return BitBoard( data & bb.data );
    }

    BitBoard operator | ( const BitBoard & bb ) const {
        return BitBoard( data | bb.data );
    }

    BitBoard operator ^ ( const BitBoard & bb ) const {
        return BitBoard( data ^ bb.data );
    }

    int operator < ( const BitBoard & bb ) const {
        return data < bb.data;
    }

    int operator > ( const BitBoard & bb ) const {
        return data > bb.data;
    }

    //
    int operator == ( const BitBoard & bb ) const {
        return data == bb.data;
    }

    int operator != ( const BitBoard & bb ) const {
        return data != bb.data;
    }

    //
    operator int () const {
        return data != 0;
    }

    //
    void setBit( unsigned idx ) {
        data |= Set[idx].data;
    }

    void clrBit( unsigned idx ) {
        data &= Clr[idx].data;
    }

    //
    int getBit( unsigned idx ) const {
        return (data & Set[idx].data) != 0 ? 1 : 0;
    }

    //
    void clear( void ) {
        data = 0;
    }

    void negate( void ) {
        data = ~data;
    }

    void andNot( const BitBoard & bb ) {
        data &= ~bb.data;
    }

    //
    unsigned getRank( int rank ) const {
        return 0xFF & (unsigned)(data >> (rank*8));
    }

    //
    int isEmpty() const {
        return (data == 0);
    }

    //
    unsigned toByte( void ) const {
        return (unsigned)data & 0xFF;
    }

    //
    unsigned toUnsigned( void ) const {
        return (unsigned)data;
    }

    // Dumps this bitboard to the logfile
    void dump( const char * header = 0 ) const;

    bool isZero() const {
        return data == 0;
    }

    bool isNotZero() const {
        return data != 0;
    }

    inline int bitScanAndResetForward() {
#if defined(USE_ASM)
#if defined(LINUX_I386) || defined(WIN_I386)
#if defined(_MSC_VER)
#pragma warning(disable: 4035)
        // Microsoft Visual C++ inline assembly for I386
        __asm {
            mov     edx, [this]
            mov     ecx, DWORD PTR [edx]
            test    ecx, ecx
             jz     High_
            bsf     eax, ecx
            dec     ecx
            and     DWORD PTR [edx], ecx
             jmp    Exit_
        High_:
            mov     ecx, DWORD PTR [edx+4]
            mov     eax, -1
            test    ecx, ecx
             jz     Exit_
            bsf     eax, ecx
            add     eax, 32
            dec     ecx
            and     DWORD PTR [edx+4], ecx
        Exit_:
        }
#elif defined(__GNUC__)
        // GCC inline assembly for I386 (note: in many years of assembly programming
        // I have never encountered a syntax so ugly and confusing... 
        // fortunately I only need short and simple code fragments here)
        int bit;

        asm(
            "   testl   %%eax, %%eax    \n\t"
            "    jz     1f              \n\t"
            "   bsf     %%eax, %0       \n\t"
            "   movl    %%eax, %%ecx    \n\t"
            "   decl    %%eax           \n\t"
            "   andl    %%ecx, %%eax    \n\t"
            "    jmp    2f              \n\t"
            "1:                         \n\t"
            "   movl    $-1, %0         \n\t"
            "   testl   %%edx, %%edx    \n\t"
            "    jz     2f              \n\t"
            "   bsf     %%edx, %0       \n\t"
            "   addl    $32, %0         \n\t"
            "   movl    %%edx, %%ecx    \n\t"
            "   decl    %%edx           \n\t"
            "   andl    %%ecx, %%edx    \n\t"
            "2:                         \n\t"
            : "=r" (bit), "=A" (data)
            : "A" (data)
            : "%ecx"
        );

        return bit;
#else
#error Unknown compiler for 32-bit I386 platform
#endif // Compiler selection
#endif // I386 platform
#else // Do not use assembly
        int bit = ::bitScanForward( data );

        if( bit >= 0 ) {
            clrBit(bit);
        }

        return bit;
#endif
    }

    int bitScanForward() const {
        return ::bitScanForward( data );
    }

    int bitCount() const {
        return ::bitCount( data );
    }

public:
    Uint64  data;
};

inline int bitScanForward( const BitBoard & b )
{
    return bitScanForward( b.data );
}

inline int bitCount( const BitBoard & b )
{
    return bitCount( b.data );
}

#endif // BITBOARD_H_
